﻿using System;
using System.Collections.Generic;

namespace Geodesic
{
	public struct Point
	{
		public Point(double lon, double lat)
		{
			Longitude = lon; Latitude = lat;
		}
		public double Longitude;
		public double Latitude;
	}
	public class Line : List<Point> { }
	public class LineSegments : List<Line> { }

	/// <summary>
	/// Extension methods for geodesic calculations.
	/// </summary>
	public static class Geodesic
	{
		/// <summary>
		/// Earth circumference in kilometers used for distance calculations - based on WGS84 ellipsoid.
		/// </summary>
		private const double EarthCircumference = 6378.137;

		/// <summary>
		/// Gets the distance between two points in Kilometers.
		/// </summary>
		/// <param name="start">The start point.</param>
		/// <param name="end">The end point.</param>
		/// <returns></returns>
		public static double GetDistanceGeodesic(Point start, Point end)
		{
			double lon1 = start.Longitude / 180 * Math.PI;
			double lon2 = end.Longitude / 180 * Math.PI;
			double lat1 = start.Latitude / 180 * Math.PI;
			double lat2 = end.Latitude / 180 * Math.PI;
			return 2 * Math.Asin(Math.Sqrt(Math.Pow((Math.Sin((lat1 - lat2) / 2)), 2) +
			 Math.Cos(lat1) * Math.Cos(lat2) * Math.Pow(Math.Sin((lon1 - lon2) / 2), 2))) * EarthCircumference;
		}

		/// <summary>
		/// Gets the shortest path line between two points. The line will be following the great
		/// circle described by the two points.
		/// </summary>
		/// <param name="start">The start point.</param>
		/// <param name="end">The end point.</param>
		/// <returns></returns>
		public static LineSegments GetGeodesicLine(Point start, Point end)
		{
			LineSegments line = new LineSegments();
			double endLongitude = end.Longitude;
			double startLongitude = start.Longitude;
			//Fix longitudes if they are outside normal +/- 180 range
			while (startLongitude < -180) startLongitude += 360;
			while (startLongitude > 180) startLongitude -= 360;
			while (endLongitude < -180) endLongitude += 360;
			while (endLongitude > 180) endLongitude -= 360;
			start = new Point(startLongitude, start.Latitude);
			end = new Point(endLongitude, end.Latitude);
			if (Math.Abs(endLongitude - startLongitude) <= 180) // Doesn't cross antimeridian 
			{
				Line pnts = GetGeodesicPoints(start, end);
				line.Add(pnts);
			}
			else
			{
				double lon1 = start.Longitude / 180 * Math.PI;
				double lon2 = end.Longitude / 180 * Math.PI;
				double lat1 = start.Latitude / 180 * Math.PI;
				double lat2 = end.Latitude / 180 * Math.PI;
				double latA = LatitudeAtLongitude(lat1, lon1, lat2, lon2, Math.PI) / Math.PI * 180; //Latitude where line crosses the antimeridian

				//One side of anti-meridian
				line.Add(GetGeodesicPoints(start, new Point(start.Longitude < 0 ? -180 : 180, latA)));
				//other side of anti-meridian
				line.Add(GetGeodesicPoints(new Point(start.Longitude < 0 ? 180 : -180, latA), end));
			}
			return line;
		}

		private static Line GetGeodesicPoints(Point start, Point end)
		{
			double lon1 = start.Longitude / 180 * Math.PI;
			double lon2 = end.Longitude / 180 * Math.PI;
			double lat1 = start.Latitude / 180 * Math.PI;
			double lat2 = end.Latitude / 180 * Math.PI;
			double dX = end.Longitude - start.Longitude;
			int points = (int)Math.Floor(Math.Abs(dX));
			dX = lon2 - lon1;
			Line pnts = new Line();
			pnts.Add(start);
			for (int i = 1; i < points; i++)
			{
				double lon = lon1 + dX / points * i;
				double lat = LatitudeAtLongitude(lat1, lon1, lat2, lon2, lon);
				pnts.Add(new Point(lon / Math.PI * 180, lat / Math.PI * 180));
			}
			pnts.Add(end);
			return pnts;
		}

		/// <summary>
		/// Gets the latitude at a specific longitude for a great circle defined by p1 and p2.
		/// </summary>
		/// <param name="p1">The p1.</param>
		/// <param name="p2">The p2.</param>
		/// <param name="lon">The longitude in degrees.</param>
		/// <returns></returns>
		private static double LatitudeAtLongitude(Point p1, Point p2, double lon)
		{
			double lon1 = p1.Longitude / 180 * Math.PI;
			double lon2 = p2.Longitude / 180 * Math.PI;
			double lat1 = p1.Latitude / 180 * Math.PI;
			double lat2 = p2.Latitude / 180 * Math.PI;
			lon = lon / 180 * Math.PI;
			return LatitudeAtLongitude(lat1, lon1, lat2, lon2, lon) / Math.PI * 180;
		}

		/// <summary>
		/// Gets the latitude at a specific longitude for a great circle defined by lat1,lon1 and lat2,lon2.
		/// </summary>
		/// <param name="lat1">The start latitude in radians.</param>
		/// <param name="lon1">The start longitude in radians.</param>
		/// <param name="lat2">The end latitude in radians.</param>
		/// <param name="lon2">The end longitude in radians.</param>
		/// <param name="lon">The longitude in radians for where the latitude is.</param>
		/// <returns></returns>
		private static double LatitudeAtLongitude(double lat1, double lon1, double lat2, double lon2, double lon)
		{
			return Math.Atan((Math.Sin(lat1) * Math.Cos(lat2) * Math.Sin(lon - lon2)
	 - Math.Sin(lat2) * Math.Cos(lat1) * Math.Sin(lon - lon1)) / (Math.Cos(lat1) * Math.Cos(lat2) * Math.Sin(lon1 - lon2)));
		}
		/// <summary>
		/// Gets the true bearing at a distance from the start point towards the new point.
		/// </summary>
		/// <param name="start">The start point.</param>
		/// <param name="end">The point to get the bearing towards.</param>
		/// <param name="distanceKM">The distance in kilometers travelled between start and end.</param>
		/// <returns></returns>
		public static double GetTrueBearingGeodesic(Point start, Point end, double distanceKM)
		{
			double d = distanceKM / EarthCircumference; //Angular distance in radians
			double lon1 = start.Longitude / 180 * Math.PI;
			double lat1 = start.Latitude / 180 * Math.PI;
			double lon2 = end.Longitude / 180 * Math.PI;
			double lat2 = end.Latitude / 180 * Math.PI;
			double tc1;
			if (Math.Sin(lon2 - lon1) < 0)
				tc1 = Math.Acos((Math.Sin(lat2) - Math.Sin(lat1) * Math.Cos(d)) / (Math.Sin(d) * Math.Cos(lat1)));
			else
				tc1 = 2 * Math.PI - Math.Acos((Math.Sin(lat2) - Math.Sin(lat1) * Math.Cos(d)) / (Math.Sin(d) * Math.Cos(lat1)));
			return tc1 / Math.PI * 180;
		}

		/// <summary>
		/// Gets the point based on a start point, a heading and a distance.
		/// </summary>
		/// <param name="start">The start.</param>
		/// <param name="distanceKM">The distance KM.</param>
		/// <param name="heading">The heading.</param>
		/// <returns></returns>
		public static Point GetPointFromHeadingGeodesic(Point start, double distanceKM, double heading)
		{
			double brng = heading / 180 * Math.PI;
			double lon1 = start.Longitude / 180 * Math.PI;
			double lat1 = start.Latitude / 180 * Math.PI;
			double dR = distanceKM / 6378.137; //Angular distance in radians
			double lat2 = Math.Asin(Math.Sin(lat1) * Math.Cos(dR) + Math.Cos(lat1) * Math.Sin(dR) * Math.Cos(brng));
			double lon2 = lon1 + Math.Atan2(Math.Sin(brng) * Math.Sin(dR) * Math.Cos(lat1), Math.Cos(dR) - Math.Sin(lat1) * Math.Sin(lat2));
			double lon = lon2 / Math.PI * 180;
			double lat = lat2 / Math.PI * 180;
			while (lon < -180) lon += 360;
			while (lat < -90) lat += 180;
			while (lon > 180) lon -= 360;
			while (lat > 90) lat -= 180;
			return new Point(lon, lat);
		}
	}
}
